const GLib = imports.gi.GLib;
const Gio = imports.gi.Gio;
const ByteArray = imports.byteArray;

/**
 * Return a Promise that resolves with the contents of a file or a default
 * value if the file doesn't exist.
 *
 * Note that if an error occurs while reading the file, the Promise may never resolve.
 *
 * @param {string} path - The path of the file to read
 * @param {string} defaultValue - The value to return if the file doesn't exist.
 *
 * @return {Promise.<string>} A Promise that resolves to the file's contents (or the default)
 */
var readFileOr = (path, defaultValue) => {
    const file = Gio.File.new_for_path(path);

    if (!file.query_exists(null)) return new Promise((resolve) => resolve(defaultValue));

    return new Promise((resolve, reject) => {
        file.load_contents_async(null, (_src, res) => {
            try {
                resolve(file.load_contents_finish(res));
            } catch (e) {
                reject(e);
            }
        });
    }).then(([loadOk, contentBytes]) => {
        if (!loadOk) {
            return defaultValue;
        }
        return ByteArray.toString(contentBytes);
    });
};

/**
 * Write a string to a file, overwriting or creating where necessary, return an (empty) Promise.
 *
 * Does not create parent folder.
 *
 * @param {string} path - The path of the file to write
 * @param {string} string - The string to write into the file
 *
 * @return {Promise} A Promise that resolves once the file has been written.
 */
var writeToFile = (path, string) => {
    const file = Gio.File.new_for_path(path);
    if (!file.query_exists(null)) {
        return new Promise((resolve, reject) => {
            file.create_async(
                Gio.FileCreateFlags.NONE,
                0, // priority 0
                null,
                (_src, res) => {
                    try {
                        resolve(file.create_finish(res));
                    } catch (e) {
                        reject(e);
                    }
                }
            );
        })
            .then(
                (outputStream) =>
                    new Promise((resolve, reject) => {
                        outputStream.write_bytes_async(
                            new GLib.Bytes(string),
                            0, // priority 0
                            null,
                            (_src, res) => {
                                try {
                                    outputStream.write_bytes_finish(res);
                                } catch (e) {
                                    reject(e);
                                }
                                resolve(outputStream);
                            }
                        );
                    })
            )
            .then(
                (outputStream) =>
                    new Promise((resolve, reject) => {
                        outputStream.close_async(
                            0, // priority 0
                            null,
                            (_src, res) => {
                                try {
                                    outputStream.close_finish(res);
                                } catch (e) {
                                    reject(e);
                                }
                                resolve();
                            }
                        );
                    })
            );
    } else {
        return new Promise((resolve, reject) => {
            file.replace_contents_bytes_async(
                new GLib.Bytes(string),
                null,
                false,
                Gio.FileCreateFlags.REPLACE_DESTINATION,
                null,
                (_src, res) => {
                    try {
                        file.replace_contents_finish(res);
                    } catch (e) {
                        reject(e);
                    }
                    resolve();
                }
            );
        });
    }
};

/**
 * List the files in a specified directory, in a Promise.
 *
 * Make sure the directory exists, or this won't resolve.
 *
 * @param {string} - The path of the directory
 *
 * @return {Promise.<string[]>} - A Promise resolving to an array of the
 *                                filenames inside the directory.
 */
var listDirectory = (path) => {
    const files = new Set();

    const dirFile = Gio.File.new_for_path(path);

    return new Promise((resolve, reject) => {
        dirFile.enumerate_children_async(
            Gio.FILE_ATTRIBUTE_STANDARD_NAME,
            Gio.FileQueryInfoFlags.NONE,
            0, // priority 0
            null,
            (_src, res) => {
                try {
                    resolve(dirFile.enumerate_children_finish(res));
                } catch (e) {
                    reject(e);
                }
            }
        );
    }).then((enumerator) => {
        let fileInfo;
        while ((fileInfo = enumerator.next_file(null))) {
            const name = fileInfo.get_attribute_as_string(Gio.FILE_ATTRIBUTE_STANDARD_NAME);
            files.add(name);
        }

        return files;
    });
};
